from abc import ABC, abstractmethod

class MacaroonService(ABC):
    """
    MacaroonService is an abstract class that defines the interface for 
    storing and retrieving macaroon related data.
    """
    @abstractmethod
    async def insert_root_key(self, token_id: bytes, root_key: bytes, macaroon: str):
        """
        Insert a new root key in the macaroon service.
        """
        pass

    @abstractmethod
    async def get_root_key(self, token_id: bytes) -> bytes:
        """
        Get the root key for the given token id.
        """
        pass


import requests
import asyncio
from l402.client import Client

from .exceptions import RequestException

class Session:
    _instance = None
    _client = None
    _configured = False

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Session, cls).__new__(cls)
        return cls._instance

    @property
    def is_configured(self) -> bool:
        """Check if the request client is configured."""
        return self._configured

    def configure(self, preimage_provider: str = None, credentials_service: str = None) -> None:
        """Configure the request client with given providers and services."""
        self._client = Client(preimage_provider, credentials_service)
        self._client.configure(preimage_provider, credentials_service)
        self._configured = True

    def request(self, method: str, url: str, **kwargs) -> requests.Response:
        """Perform a request with optional parameters."""
        if not self.is_configured:
            raise RequestException("No request client configured.")
        return asyncio.run(self._client.request(method, url, **kwargs))

    def get(self, url: str, **kwargs) -> requests.Response:
        """Perform a GET request with optional parameters."""
        return self.request('GET', url, **kwargs)

    def post(self, url: str, **kwargs) -> requests.Response:
        """Perform a POST request with optional parameters."""
        return self.request('POST', url, **kwargs)

    def put(self, url: str, **kwargs) -> requests.Response:
        """Perform a PUT request with optional parameters."""
        return self.request('PUT', url, **kwargs)

    def delete(self, url: str, **kwargs) -> requests.Response:
        """Perform a DELETE request with optional parameters."""
        return self.request('DELETE', url, **kwargs)





from .authenticator import Authenticator
from .invoice_provider import InvoiceProvider
from .macaroons import MacaroonService
from .exceptions import InvalidOrMissingL402Header, InvalidMacaroon
from .middlewares import Flask_l402_decorator, FastAPIL402Middleware


from abc import ABC, abstractmethod

class PreimageProvider(ABC):
    """
    Abstract Base Class (ABC) for a Preimage Provider.

    A Preimage Provider is responsible for retrieving the preimage associated with a given invoice.
    Any concrete class that inherits from this ABC must implement the `get_preimage` method.
    """

    @abstractmethod
    async def get_preimage(self, invoice: str) -> str:
        """
        Abstract method to retrieve the preimage for a given invoice.

        Args:
            invoice (str): The unique identifier of the invoice for which the preimage is to be retrieved.

        Returns:
            str: The preimage associated with the given invoice. The preimage is a string that, when hashed,
                 produces the hash embedded in the invoice.

        Raises:
            NotImplementedError: This method must be implemented by any concrete class that inherits from this ABC.
        """
        pass


import httpx
import asyncio

from .preimage_provider import PreimageProvider
from .credentials import CredentialsService, parse_http_402_response

class Client:
    """
    The Client class is a low-level HTTP client implementation that handles HTTP requests 
    with 402 Payment Required responses.
    """

    def __init__(self, preimage_provider: PreimageProvider = None, 
                 credentials_service: CredentialsService = None):
        self._preimage_provider = preimage_provider
        self._credentials_service = credentials_service

        self._lock = asyncio.Lock()

    @property
    def preimage_provider(self) -> PreimageProvider:
        return self._preimage_provider

    @property
    def credentials_service(self) -> CredentialsService:
        return self._credentials_service

    def configure(self, preimage_provider: PreimageProvider = None, credentials_service: CredentialsService = None):
        """Configures the client with the given services."""
        if preimage_provider:
            self._preimage_provider = preimage_provider
        if credentials_service:
            self._credentials_service = credentials_service

    def _add_authorization_header(self, kwargs, credentials):
        """Adds the L402 Authorization header to the request."""
        headers = kwargs.setdefault('headers', {})
        headers['Authorization'] = credentials.authentication_header()

    async def _handle_402_payment_required(self, url: str, response: httpx.Response) -> CredentialsService:
        """Handles a 402 Payment Required response."""
        creds = parse_http_402_response(response)
        creds.set_location(url)
        preimage = await self.preimage_provider.get_preimage(creds.invoice)
        if not preimage:
            raise Exception("Payment failed.")

        creds.preimage = preimage
        await self.credentials_service.store(creds)
        return creds

    async def request(self, method: str, url: str, **kwargs) -> httpx.Response:
        async with self._lock:
            creds = await self.credentials_service.get(url)
            if creds:
                self._add_authorization_header(kwargs, creds)

            async with httpx.AsyncClient(timeout=30.0) as client:
                response = await client.request(method, url, **kwargs)
                if response.status_code != 402:
                    return response

                new_creds = await self._handle_402_payment_required(url, response)
                self._add_authorization_header(kwargs, new_creds)
                return await client.request(method, url, **kwargs)



class InvalidOrMissingL402Header(Exception):
    """Exception raised for errors in the L402 header format."""
    pass

class InvalidMacaroon(Exception):
    """Exception raised for errors when validating macaroons."""
    pass


from abc import ABC, abstractmethod

from .credentials import L402Credentials

class CredentialsService(ABC):
    """
    Abstract Base Class (ABC) for a Credentials Service.

    A Credentials Service is responsible for storing and retrieving L402 Credentials.
    Any concrete class that inherits from this ABC must implement the `store` and `get` methods.
    """

    @abstractmethod
    async def store(self, credentials: L402Credentials):
        """
        Abstract method to store a new L402Credentials in the credentials service.

        Args:
            credentials (L402Credentials): The credentials to be stored.

        Raises:
            NotImplementedError: This method must be implemented by any concrete class that inherits from this ABC.
        """
        pass

    @abstractmethod
    async def get(self, location: str):
        """
        Abstract method to retrieve the L402Credentials object for a given location.

        Args:
            location (str): The location for which the credentials are to be retrieved.

        Returns:
            L402Credentials: The credentials associated with the given location.

        Raises:
            NotImplementedError: This method must be implemented by any concrete class that inherits from this ABC.
        """
        pass


from .client import Client
from .credentials import L402Credentials, CredentialsService, parse_http_402_response
from .preimage_provider import PreimageProvider
from .requests import Session

# Create the singleton instance
requests = Session()


import httpx
import json
from typing import Tuple

from .invoice_provider import InvoiceProvider

class AlbyAPI(InvoiceProvider):
    def __init__(self, api_key):
        self.api_key = api_key
        self.alby_url = "https://api.getalby.com"
    
    def _prepare_request(self, amount: int, currency: str, description: str) -> Tuple[str, str]:
        url = f"{self.alby_url}/invoices"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        data = json.dumps({
            "amount": amount, 
            "currency": currency,
            "description": description,
        })
        return url, headers, data
    
    def _process_response(self, response):
        response.raise_for_status()

        try:
            new_invoice_response = response.json()
        except json.JSONDecodeError:
            raise Exception(f"Invalid JSON response: {response.text}")
        
        payment_request = new_invoice_response.get("payment_request")
        payment_hash = new_invoice_response.get("payment_hash")

        if not payment_request or not payment_hash:
            raise Exception(f"No payment request or hash found in response: {new_invoice_response}")
        
        return payment_request, payment_hash
    
    async def create_invoice(self, amount: int, currency: str, description: str) -> Tuple[str, str]:
        url, headers, data = self._prepare_request(amount, currency, description)
        async with httpx.AsyncClient(timeout=30.0) as client:
            response = await client.post(url, headers=headers, content=data)
            return self._process_response(response)
    


from typing import Tuple
from abc import ABC, abstractmethod

class InvoiceProvider(ABC):
    """Abstract Base Class (ABC) for an Invoice Provider.

    This abstract base class defines the interface for classes that provide Lightning invoices.
    Any concrete class that inherits from this ABC must implement its abstract methods.
    """

    @abstractmethod
    async def create_invoice(self, amount: int, currency: str, description: str) -> Tuple[str, str]:
        """Abstract method to create a new invoice with a given amount, currency, and description.

        Args:
            amount (int): The amount of the invoice.
            currency (str): The currency of the invoice.
            description (str): A brief description of the purpose of the invoice.

        Returns:
            Tuple[str, str]: A tuple containing the payment_request and payment_hash of the invoice.

        Raises:
            NotImplementedError: This method must be implemented by any concrete class that inherits from this ABC.
        """
        pass



import os
import re
import hashlib
import struct
from typing import Tuple

from binascii import hexlify, unhexlify
from pymacaroons import Macaroon, Verifier, MACAROON_V2

from .invoice_provider import InvoiceProvider
from .macaroons import MacaroonService
from .exceptions import InvalidOrMissingL402Header, InvalidMacaroon
    
# Parse the L402 header pattern: "L402 <macaroon>:<preimage>"
L402_HEADER_PATTERN = re.compile(r'^L402\s+(.*?):(.*?)$')


class Authenticator:
    """
    The Authenticator class implements the L402 authentication protocol. 
    
    It can be used to generate new challenges for requests that do not include an L402 header,
    and also validate the L402 headers in the incoming requests.
    """
    def __init__(self, location: str, invoice_provider: InvoiceProvider, macaroon_service: MacaroonService):
        self.location = location
        self.invoice_provider = invoice_provider
        self.macaroon_service = macaroon_service

    async def new_challenge(self, amount: int, currency: str, description: str) -> Tuple[str, str]:
        """Generate a new L402 challenge with a new macaroon and invoice."""
        # Create a new invoice
        payment_request, payment_hash  = await self.invoice_provider.create_invoice(
            amount, currency, f"L402 Challenge: {description}",
        )
        
        # Generate new revoking keys for the macaroon
        token_id = os.urandom(32)
        root_key = os.urandom(32)

        identifier = self._encode_identifier(0, payment_hash, token_id)

        # Generate a new macaroon with the root key
        mac = Macaroon(
            version=MACAROON_V2, 
            location=self.location,
            identifier=identifier, 
            key=root_key,
        )

        # TODO(positiveblue): Add support for custom caveats.

        encoded_macaroon = mac.serialize()
        await self.macaroon_service.insert_root_key(token_id, root_key, encoded_macaroon)

        return encoded_macaroon, payment_request

    async def validate_l402_header(self, header: str):
        """Validate the L402 header and its contents."""
        encoded_macaroon, preimage = self._parse_l402_header(header)
        mac, payment_hash, token_id = self._decode_macaroon(encoded_macaroon)

        self._validate_preimage(preimage, payment_hash)
        await self._validate_macaroon(mac, token_id)
        await self._validate_caveats(mac)

    def _encode_identifier(self, version, payment_hash, token_id):
        """Encode the L402 identifier."""
        payment_hash_bytes = unhexlify(payment_hash)
        return struct.pack(">H32s32s", version, payment_hash_bytes, token_id)
    
    def _decode_identifier(self, identifier):
        """Decode the L402 identifier."""
        version, payment_hash, token_id = struct.unpack(">H32s32s", identifier)
        payment_hash_hex = hexlify(payment_hash).decode()

        return version, payment_hash_hex, token_id

    def _parse_l402_header(self, header):
        """Parse the L402 header and return the encoded macaroon and preimage."""
        match = L402_HEADER_PATTERN.match(header)
        if not match:
            raise InvalidOrMissingL402Header(f"Invalid L402 header format: {header}")
        
        encoded_macaroon, preimage = match.groups()
        if not encoded_macaroon.strip() or not preimage.strip():
            raise InvalidOrMissingL402Header(f"Macaroon or preimage cannot be empty: {header}")
        
        return encoded_macaroon, preimage

    def _decode_macaroon(self, encoded_macaroon):
        """Return the relevant L402 information from the encoded macaroon."""
        # Deserialize the macaroon
        mac = Macaroon.deserialize(encoded_macaroon)

        # Check the version in the macaroon identifier
        version, payment_hash, token_id = self._decode_identifier(mac.identifier)
        if version != 0:
            raise ValueError(f"Invalid L402 version: {version}")
        
        return mac, payment_hash, token_id
    
    def _validate_preimage(self, preimage: str, payment_hash: str):
        """Validate the macaoon hash preimage."""
        preimage_bytes = unhexlify(preimage)
        computed_hash = hashlib.sha256(preimage_bytes).digest()

        payment_hash_bytes = bytes.fromhex(payment_hash)

        if payment_hash_bytes != computed_hash:
            raise ValueError("Invalid payment preimage")
    
    async def _validate_macaroon(self, mac, token_id):
        """Verify the macaroon with the linked root key."""
        root_key = await self.macaroon_service.get_root_key(token_id)
        
        verifier = Verifier()
        try:
            verifier.verify(mac, root_key)
        except Exception:
            raise InvalidMacaroon("Macaroon verification failed.")
    
    async def _validate_caveats(self, mac):
        """Validate the macaroon caveats."""
        # TODO(positiveblue): Implement custom caveats validation
        pass


import os
import sqlite3
from datetime import datetime

from .macaroon_service import MacaroonService

def adapt_datetime(dt):
    return dt.isoformat()

# Register the custom datetime adapter
# This is needed because the default datetime adapter is deprecated in Python 3.12 and later versions
# The custom adapter converts datetime objects to ISO 8601 string format for storage in the database
sqlite3.register_adapter(datetime, adapt_datetime)

class SqliteMacaroonService(MacaroonService):
    """
    SqliteMacaroonService is an SQLite-based credentials service for L402.
    """

    def __init__(self, path=None):
        self.db_path = path or os.path.join(os.path.expanduser('~'), 'authenticator.db')
        self.conn = sqlite3.connect(self.db_path)
        self._create_table()

    def _create_table(self):
        create_table_sql = """
            CREATE TABLE IF NOT EXISTS macaroons (
                -- id is the primary key of the table.
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                
                -- token_id is ...
                token_id BLOB UNIQUE NOT NULL,

                -- root_key is...
                root_key BLOB NOT NULL,
                        
                -- macaroon is the "admin" base64 encoded macaroon
                macaroon TEXT,

                -- created_at is the date and time when the credentials were
                -- created.
                created_at DATETIME NOT NULL
            );
            
            CREATE INDEX IF NOT EXISTS macaroons_token_id_idx ON macaroons (token_id);
        """
        cursor = self.conn.cursor()
        cursor.executescript(create_table_sql)
        self.conn.commit()
    
    async def insert_root_key(self, token_id: bytes, root_key: bytes, macaroon: str):
        insert_sql = """
            INSERT INTO macaroons (
                token_id, root_key, macaroon, created_at
            ) VALUES (
                ?, ?, ?, ?
            );
        """

        created_at = datetime.now()

        cursor = self.conn.cursor()
        cursor.execute(
            (insert_sql), 
            (token_id, root_key, macaroon, created_at),
        )
        self.conn.commit()
    
    async def get_root_key(self, token_id: bytes) -> bytes:
        query_sql = """
            SELECT root_key
            FROM macaroons
            WHERE token_id = ?
        """

        cursor = self.conn.cursor()
        cursor.execute(
            query_sql, (token_id,)
        )

        row = cursor.fetchone()
        if row is None:
            return None

        return row[0]
        
    def __del__(self):
        """
        Close the SQLite connection.
        """
        self.conn.close()


from typing import Callable, Tuple, Coroutine, Any
import asyncio

from fastapi import Request, HTTPException, Response
from starlette.middleware.base import BaseHTTPMiddleware
from functools import wraps
from flask import request, make_response, current_app

from l402.server import Authenticator

class FastAPIL402Middleware(BaseHTTPMiddleware):
    def __init__(
        self, 
        app,
        authenticator: Authenticator, 
        pricing_func: Callable[[Request], Tuple[str, str, str]],
    ):
        super().__init__(app)
        self.authenticator = authenticator
        self.pricing_func = pricing_func

    async def dispatch(
        self, 
        request: Request, 
        call_next: Callable[[Request], Coroutine[Any, Any, Response]],
    ) -> Response:

        header = request.headers.get("Authorization")
        if header:
            try:
                await self.authenticator.validate_l402_header(header)
                return await call_next(request)
            except Exception as e:
                pass
        
        amount, currency, description = self.pricing_func(request)
        macaroon, payment_request = await self.authenticator.new_challenge(amount, currency, description)
        response = HTTPException(status_code=402, detail="Payment Required")
        response.headers["WWW-Authenticate"] = f'L402 macaroon="{macaroon}", invoice="{payment_request}"'
        raise response


def Flask_l402_decorator(authenticator, pricing_func):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            async def async_wrapper():
                with current_app.app_context():
                    header = request.headers.get("Authorization")
                    if header:
                        try:
                            await authenticator.validate_l402_header(header)
                            if asyncio.iscoroutinefunction(func):
                                response = await func(*args, **kwargs)
                            else:
                                response = func(*args, **kwargs)
                            return response
                        
                        except Exception as e:
                            pass

                    amount, currency, description = pricing_func(request)
                    macaroon, payment_request = await authenticator.new_challenge(amount, currency, description)
                    response = make_response("Payment Required", 402)
                    response.headers["WWW-Authenticate"] = f'L402 macaroon="{macaroon}", invoice="{payment_request}"'
                    return response

            return asyncio.run(async_wrapper())

        return wrapper

    return decorator


import os
import sqlite3
from datetime import datetime

from .credentials import L402Credentials
from .credentias_service import CredentialsService

def adapt_datetime(dt):
    return dt.isoformat()

# Register the custom datetime adapter
# This is needed because the default datetime adapter is deprecated in Python 3.12 and later versions
# The custom adapter converts datetime objects to ISO 8601 string format for storage in the database
sqlite3.register_adapter(datetime, adapt_datetime)

class SqliteCredentialsService(CredentialsService):
    """
    SqliteCredentialsService is an SQLite-based credentials service for L402.
    """

    def __init__(self, path=None):
        self.db_path = path or os.path.join(os.path.expanduser('~'), 'credentials.db')
        self.conn = sqlite3.connect(self.db_path)
        self._create_table()

    def _create_table(self):
        create_table_sql = """
            CREATE TABLE IF NOT EXISTS credentials (
                -- id is the primary key of the table.
                id INTEGER PRIMARY KEY AUTOINCREMENT,

                -- location is the url for the resource.
                location TEXT NOT NULL,

                -- macaroon is the base64 encoded macaroon needed in the
                -- L402 request header.
                macaroon TEXT NOT NULL,

                -- preimage is the preimage linked to the macaroon payment
                -- hash. Also needed in the L402 request header.
                preimage TEXT,

                -- invoice is the LN invoice that was paid to complete the
                -- credentials.
                invoice TEXT NOT NULL,

                -- created_at is the date and time when the credentials were
                -- created.
                created_at DATETIME NOT NULL
            );
            
            CREATE INDEX IF NOT EXISTS credentials_location_index ON credentials (location);
        """
        cursor = self.conn.cursor()
        cursor.executescript(create_table_sql)
        self.conn.commit()
    
    async def store(self, credentials: L402Credentials):
        insert_sql = """
            INSERT INTO credentials (
                location, macaroon, preimage, invoice, created_at
            ) VALUES (
                ?, ?, ?, ?, ?
            );
        """

        location = credentials.location
        macaroon = credentials.macaroon
        preimage = credentials.preimage
        invoice = credentials.invoice
        created_at = datetime.now()

        cursor = self.conn.cursor()
        cursor.execute(
            (insert_sql), 
            (location, macaroon, preimage, invoice, created_at),
        )
        self.conn.commit()
    
    async def get(self, location: str):
        query_sql = """
            SELECT macaroon, preimage, invoice
            FROM credentials
            WHERE location = ?
            ORDER BY created_at DESC
            LIMIT 1
        """

        cursor = self.conn.cursor()
        cursor.execute(
            query_sql, (location,)
        )

        row = cursor.fetchone()
        if row:
            macaroon, preimage, invoice = row
            credentials = L402Credentials(macaroon, preimage, invoice)
            credentials.set_location(location)
            return credentials
        
        return None

    def __del__(self):
        """
        Close the SQLite connection.
        """
        self.conn.close()


import httpx

from .client import Client
from .preimage_provider import PreimageProvider
from .credentials import CredentialsService, parse_http_402_response

_default_preimage_provider = None
_default_credentials_service = None

def configure(preimage_provider: PreimageProvider, credentials_service: CredentialsService):
    global _default_preimage_provider
    global _default_credentials_service

    _default_preimage_provider = preimage_provider
    _default_credentials_service = credentials_service
    
class AsyncClient(httpx.AsyncClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if _default_preimage_provider is None or _default_credentials_service is None:
            raise Exception("You must configure the client before using it.")
        
        self._preimage_provider = _default_preimage_provider
        self._credentials_service = _default_credentials_service
    
    def _add_authorization_header(self, request, credentials):
        """Adds the L402 Authorization header to the request."""
        request.headers['Authorization'] = credentials.authentication_header()

    async def _handle_402_payment_required(self, url: str, response: httpx.Response) -> CredentialsService:
        """Handles a 402 Payment Required response."""
        creds = parse_http_402_response(response)
        creds.set_location(url)
        preimage = await self._preimage_provider.get_preimage(creds.invoice)
        if not preimage:
            raise Exception("Payment failed.")

        creds.preimage = preimage
        await self._credentials_service.store(creds)

        return creds
    
    async def send(self, request, *args, **kwargs):
        url = str(request.url)
        creds = await self._credentials_service.get(url)
        if creds:
            self._add_authorization_header(request, creds)

        response = await super().send(request, *args, **kwargs)
        if response.status_code != 402:
            return response

        new_creds = await self._handle_402_payment_required(url, response)
        self._add_authorization_header(request, new_creds)
        return await super().send(request, *args, **kwargs)
    



import os
import httpx
import json
from typing import Tuple
from .invoice_provider import InvoiceProvider

class FewsatsInvoiceProvider(InvoiceProvider):
    """Concrete implementation of InvoiceProvider using the Fewsats API."""

    def __init__(self, base_url: str = "https://api.fewsats.com", api_key: str = ""):
        """Initialize the FewsatsInvoiceProvider.

        Args:
            base_url (str): The base URL of the Fewsats API. Defaults to "https://api.fewsats.com".
        """
        self.base_url = base_url
        if api_key:
            self.api_key = api_key
        else:
            self.api_key = os.environ.get("FEWSATS_API_KEY")
        if not self.api_key:
            raise ValueError("Missing FEWSATS_API_KEY environment variable. Get one at https://fewsats.com/")

    def _prepare_request(self, amount: int, currency: str, description: str) -> Tuple[str, dict, str]:
        url = f"{self.base_url}/v0/invoices"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}"
        }
        data = json.dumps({
            "amount": amount,
            "currency": currency,
            "description": description
        })
        return url, headers, data

    def _process_response(self, response: httpx.Response) -> Tuple[str, str]:
        response.raise_for_status()

        try:
            invoice_data = response.json()
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON response: {response.text}")

        payment_request = invoice_data.get("PaymentRequest")
        payment_hash = invoice_data.get("PaymentHash")

        if not payment_request or not payment_hash:
            raise ValueError(f"Invalid response from Fewsats API: {invoice_data}")

        return payment_request, payment_hash

    async def create_invoice(self, amount: int, currency: str, description: str) -> Tuple[str, str]:
        """Create a new invoice using the Fewsats API.

        Args:
            amount (int): The amount of the invoice in the smallest unit of the currency (e.g., cents for USD).
            currency (str): The currency of the invoice (e.g., "USD").
            description (str): A brief description of the purpose of the invoice.

        Returns:
            Tuple[str, str]: A tuple containing the payment_request and payment_hash of the invoice.

        Raises:
            ValueError: If the API request fails or returns an unexpected response.
        """
        url, headers, data = self._prepare_request(amount, currency, description)
        async with httpx.AsyncClient(timeout=30.0) as client:
            response = await client.post(url, headers=headers, content=data)
            return self._process_response(response)



from .credentials import L402Credentials, parse_http_402_response, _parse_l402_challenge
from .sqlite_credentials_service import SqliteCredentialsService
from .credentias_service import CredentialsService


from typing import Optional, Union
import re

from httpx import Response as HTTPXResponse
from requests import Response as RequestsResponse

HTTPResponse = Union[RequestsResponse, HTTPXResponse]

MACAROON_REGEX = re.compile(r'macaroon="([^ ]+)"')
INVOICE_REGEX = re.compile(r'invoice="([^ ]+)"')

class L402Credentials:
    """
    A class to represent the credentials required to authenticate a user trying
    to access an L402 protected resource.
    """
    def __init__(self, macaroon: str, preimage: Optional[str], invoice: str):
        self.macaroon = macaroon
        self.preimage = preimage
        self.invoice = invoice

    def authentication_header(self) -> str:
       return f"L402 {self.macaroon}:{self.preimage}"
    
    def set_location(self, location: str):
        # TODO(positiveblue): we need something like "policies". Policies will 
        # allow us to define the conditions under which a credential can be used.
        # By now we will simply set the url.
        self.location = location


def parse_http_402_response(response: HTTPResponse) -> L402Credentials:
    """
    Parse the L402 challenge from a http response with an 402 status code.
    """

    # TODO(positiveblue): check that fastAPI does not change the case of 
    # headers to lowercase.
    challenge = response.headers.get('WWW-Authenticate')
    if not challenge:
        raise ValueError("WWW-Authenticate header missing.")
    
    return _parse_l402_challenge(challenge)


def _parse_l402_challenge(challenge: str) -> L402Credentials:
    """
    Parse an L402 challenge string. Currently the V0 challenge format is:
        `L402 macaroon="{macaroon}", invoice="{invoice}"`
    """
    # TODO(positiveblue): we should check that it has the right format but for 
    # now we will simply patter match the macaroon and invoice.
    macaroon_match = MACAROON_REGEX.search(challenge)
    invoice_match = INVOICE_REGEX.search(challenge)

    if not macaroon_match or not invoice_match:
        raise ValueError(
            f"Challenge parsing failed, macaroon or invoice missing. Challenge: {challenge}",
        )
    
    macaroon = macaroon_match.group(1)
    invoice = invoice_match.group(1)

    # The preimage is currently unkown so we will set it to None.
    preimage = None

    return L402Credentials(macaroon, preimage, invoice)


from typing import Tuple, Dict

import httpx
import json

from .preimage_provider import PreimageProvider

class AlbyAPI(PreimageProvider):
    def __init__(self, api_key: str):
        self.api_key = api_key
        self.alby_url = "https://api.getalby.com"

    async def get_preimage(self, invoice: str) -> str:
        """
        Retrieves the preimage for the given invoice.

        Args:
            invoice (str): The invoice for which to retrieve the preimage.

        Returns:
            str: The preimage associated with the invoice.
        
        Raises:
            Exception: If unable to obtain the preimage.
        """
        url, headers, data = self._prepare_request(invoice)
        async with httpx.AsyncClient(timeout=30.0) as client:
            response = await client.post(url, headers=headers, content=data)
            return self._process_response(response)

    def _prepare_request(self, invoice: str) -> Tuple[str, str, str]:
        """
        Prepares the request to the Alby API to retrieve the preimage for the given invoice.

        Args:
            invoice (str): The invoice for which to retrieve the preimage.

        Returns:
            Tuple[str, str, str]: The URL, headers, and data for the request.
        """
        url = f"{self.alby_url}/payments/bolt11"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        data = json.dumps({'invoice': invoice})
        return url, headers, data

    def _process_response(self, response: httpx.Response) -> str:
        """
        Processes the response from the Alby API to retrieve the preimage.
        
        Args:
            response (httpx.Response): The response from the Alby API.
            
        Returns:
            str: The preimage associated with the invoice.
                
        Raises:
            Exception: If the response status code is not 200.
            Exception: If the response is not valid JSON with a payment preimage.
        """
        if response.status_code != 200:
            raise Exception(f"Unexpected response {response.status_code}: {response.text}, headers: {response.headers}, request details: {response.request}")


        try:
            payment_response = response.json()
        except json.JSONDecodeError:
            raise Exception(f"Invalid JSON response: {response.text}")
        
        preimage = payment_response.get("payment_preimage")
        if not preimage:
            raise Exception(f"Payment preimage not found in response: {payment_response}")

        return preimage